import json
import re
from ruly import common


def parse(rule_str):
    """Parses a string representation of a rule and returns it

    Args:
        rule_str (str): string representation of a rule

    Returns:
        ruly.Rule"""

    raise NotImplementedError()
    if_antecedent_str, consequent_str = rule_str.split('THEN')
    antecedent_str = if_antecedent_str.split('IF')[1]
    return common.Rule(_parse_antecedent(antecedent_str),
                       _parse_consequent(consequent_str))


def rule_str(rule):
    """Returns a human-readable rule representation

    Args:
        rule (ruly.Rule): rule

    Returns:
        str"""
    return (f'IF {_repr_antecedent(rule.antecedent)} '
            f'THEN {_repr_consequent(rule.consequent)}')


def _parse_antecedent(string):
    raise NotImplementedError()
    string = string.strip()
    parenth_regex = r'\((.+)\)'
    subexpressions = re.findall(parenth_regex, string)

    def repl(m):
        for i, _ in enumerate(subexpressions):
            yield f'_subexpression{i}'
    string = re.sub(parenth_regex, repl, string)

    op_locations = [(op, string.find(op.name)) for op in common.Operator]
    op = min([(op, l) for op, l in op_locations if l != -1],
             key=lambda tup: tup[1], default=None)
    if op is None:
        if string == '_subexpression0':
            return _parse_antecedent(subexpressions[0])

        # return _parse_condition(string)
        return None
    children = []
    for child_str in string.split(op.name):
        if '_subexpression' in child_str:
            subexp_index = int(child_str.split('_subexpression')[1])
            child_str = subexpressions[subexp_index]
        children.append(_parse_antecedent(child_str))
    return common.Expression(op, children)


def _parse_consequent(string):
    return common.Assignment(sub.strip() for sub in string.split('='))


def _repr_antecedent(antecedent):
    if isinstance(antecedent, common.Expression):
        return _repr_expression(antecedent)
    return repr(antecedent)


def _repr_expression(expression):
    child_reprs = []
    for child in expression.children:
        if isinstance(child, common.Expression):
            child_reprs.append(f'({_repr_expression(child)})')
        elif isinstance(child, common.Condition):
            child_reprs.append(repr(child))
    return f' {expression.operator.name} '.join(child_reprs)


def _repr_consequent(assignment):
    return f'{assignment.name} = {json.dumps(assignment.value)}'



from ruly import common


def backward_chain(knowledge_base, output_name, post_eval_cb=None, **kwargs):
    """Evaulates the output using backward chaining

    The algorithm is depth-first-search, if goal variable assigment is
    contained within a rule that has a depending derived variable, this
    variable is solved for using the same function call.

    Args:
        knowledge_base (ruly.KnowledgeBase): knowledge base
        output_name (str): name of the goal variable
        post_eval_cb(Optional[Callable]): callback called after determining
            which rules fired, signature should match :func:`post_eval_cb`.
            Return value is changed state. If `None`, state is changed by using
            assignemnt of first fired rule's consequent (or not changed if no
            rules fired)
        **kwargs (Dict[str, Any]): names and values of input variables

    Returns:
        Dict[str, Any]: state containing calculated values
    """
    state = {
        name: kwargs.get(name)
        for name in knowledge_base.input_variables.union(
            knowledge_base.derived_variables)}
    if state[output_name] is not None:
        return state

    fired_rules = []
    for rule in [r for r in knowledge_base.rules
                 if output_name in r.consequent]:
        depending_variables = common.get_rule_depending_variables(rule)
        for depending_variable in [var for var in depending_variables
                                   if state[var] is None]:
            if depending_variable in knowledge_base.input_variables:
                break
            eval_state = backward_chain(knowledge_base, depending_variable,
                                        post_eval_cb=post_eval_cb, **state)
            state = dict(state, **eval_state)
            if state[depending_variable] is None:
                break
        if evaluate(state, rule.antecedent):
            if post_eval_cb is None:
                return dict(state, **rule.consequent)
            fired_rules.append(rule)

    if post_eval_cb:
        return post_eval_cb(state, output_name, fired_rules)
    return state


def post_eval_cb(state, output_name, fired_rules):
    """Placeholder function describing input and output arguments for the post
    evaluation callbacks

    Args:
        state (Dict[str, Any]): state calculated during evaluation
        output_name (str): name of the goal variable
        fired_rules (List[ruly.Rule]): rules whose antecedents were satisfied
            during the evaluation
    Returns:
        Dict[str, Any]: updated state"""


def evaluate(inputs, antecedent):
    """Evaluates truthiness of an antecedent

    Args:
        inputs (Dict[str, Any]): variable values
        antecedent (Union[ruly.Expression, ruly.Condition]): rule
            antecedent

    Returns:
        bool"""
    if isinstance(antecedent, common.Condition):
        return _evaluate_condition(antecedent, inputs[antecedent.name])
    elif isinstance(antecedent, common.Expression):
        return _evaluate_expression(antecedent, inputs)


def _evaluate_expression(expression, inputs):
    if expression.operator == common.Operator.AND:
        return all([evaluate(inputs, child) for child in expression.children])


def _evaluate_condition(condition, input_value):
    if isinstance(condition, common.EqualsCondition):
        return condition.value == input_value



from ruly.common import (Rule,
                         Operator,
                         Expression,
                         Condition,
                         EqualsCondition,
                         Evaluation,
                         get_rule_depending_variables)
from ruly.evaluator import (backward_chain,
                            post_eval_cb,
                            evaluate)
from ruly.knowledge_base import KnowledgeBase
from ruly.parser import parse


__all__ = ['Rule',
           'Operator',
           'Expression',
           'Condition',
           'EqualsCondition',
           'Evaluation',
           'get_rule_depending_variables',
           'backward_chain',
           'post_eval_cb',
           'evaluate',
           'KnowledgeBase',
           'parse']



from ruly import (common, parser)


class KnowledgeBase:
    """Class representing a knowledge base, stores known rules

    Args:
        *args (Tuple[Union[ruly.Rule, ruly.KnowledgeBase, str]]): initial rules
            of which the knowledge base consists, can be rule objects,
            knowledge bases or string representations
        parser(Callable[[str], ruly.Rule]): parser used to turn string
            representations into ruly.Rule objects, default is
            :meth:`ruly.parse`
    """

    def __init__(self, *args, parse=parser.parse):
        input_variables = set()
        derived_variables = set()
        rules = []
        for arg in args:
            if isinstance(arg, str):
                arg = self._parse(arg)
            elif isinstance(arg, KnowledgeBase):
                rules.extend(arg.rules)
            elif isinstance(arg, common.Rule):
                rules.append(arg)
            else:
                raise ValueError()
        for rule in rules:
            input_variables.update(common.get_rule_depending_variables(rule))
            derived_variables.update(rule.consequent)
            input_variables -= derived_variables

        self._rules = tuple(rules)
        self._input_variables = input_variables
        self._derived_variables = derived_variables
        self._parse = parse

    @property
    def rules(self):
        """List[ruly.Rule]: Stored rules"""
        return self._rules

    @property
    def input_variables(self):
        """Set[str]: Names of variables that are never contained within a
            rule's antecedent"""
        return self._input_variables

    @property
    def derived_variables(self):
        """Set[str]: Names of variables contained within at least one rule's
            antecedent"""
        return self._derived_variables



import setuptools

import ruly.version


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='ruly',
    packages=['ruly'],
    version=ruly.version.version,
    url='https://github.com/ZlatSic/ruly',
    author='Zlatan Sičanica',
    author_email='zlatan.sicanica@gmail.com',
    description='Simple, extensible rule engine in Python',
    long_description=long_description,
    long_description_content_type='text/markdown')
